#coding=utf-8

import hashlib

#计算文件hash值
def fhash(fp, hm="sha256", blk_sz = 10240):
	if type(hm) == str:
		hm = getattr(hashlib, hm)()
	with open(fp, 'rb') as f:
		while True:
			bs = f.read(blk_sz)
			if len(bs)==0:
				break
			hm.update(bs)
	return hm.hexdigest()

pass



#
from ..ioc.base import Base, EncapeData,IOCError
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
class ObjectDeal(FormatDeal):
    """
        对象object:
            {
                id: id
                type: object
                source: 导入路径+调用方法/类
                single: 1 //是否单例，默认是，
                        //这里的单例是一个id对应一个实例，
                        //如果两个id用的同一个source，就是同一个source的两个对象
                // 构造函数
                construct:{
                    args: [
                        item_conf,
                        ...
                    ]
                    maps: {
                        key1: item_conf,
                        ...
                    }
                }
                //construct和args+maps，不能同时存在
                args: [
                    item_conf,
                    ...
                ]
                maps: {
                    key1: item_conf,
                    ...
                }
                // sets之前调用方法
                prev_call: item_conf
                // 对象变量设置属性
                sets: [
                    {key: key1, data: item_conf }
                    ...
                ]
                // sets之后调用
                call: item_conf
                // remove前调用
                remove: item_conf
                // remove后调用
                after_remove: item_conf
            }
        简写：
            [[id, object, single], source, construct, sets, calls] 
                construct: [args, maps]
                sets: [[key1, item_conf], ...]
        极简:
            [object, source]
        例:
            [object, buildz.ioc.ioc.conf.Conf] //生成Conf对象
    """
    def init(self, fp_lists = None, fp_defaults = None, fp_cst = None, fp_set = None):
        self.singles = {}
        self.sources = {}
        super().init("ObjectDeal", fp_lists, fp_defaults, 
            join(dp, "conf", "obj_lists.js"),
            join(dp, "conf", "obj_defaults.js"))
        if fp_set is None:
            fp_set = join(dp, 'conf', 'obj_set_lists.js')
        if fp_cst is None:
            fp_cst = join(dp, 'conf', 'obj_cst_lists.js')
        self.fmt_set = FormatData(xf.loads(xf.fread(fp_set)))
        self.fmt_cst = FormatData(xf.loads(xf.fread(fp_cst)))
    def get_maps(self, maps, sid, id):
        if id is None:
            return None
        if sid not in maps:
            return None
        maps = maps[sid]
        if id not in maps:
            return None
        return maps[id]
    def set_maps(self, maps, sid, id, obj):
        if sid not in maps:
            maps[sid] = {}
        maps[sid][id] = obj
    def deal(self, edata:EncapeData):
        sid = edata.sid
        data = edata.data
        data = self.format(data)
        info = edata.info
        conf = edata.conf
        confs = edata.confs
        if type(info) == dict:
            cid = xf.g(info, id=None)
        else:
            cid = None
        id = xf.g(data, id = None)
        single = xf.g(data, single=None)
        if id is None:
            single = 0
        if single is None:
            single = 1
        ids = None
        if single or cid is not None:
            if cid is not None:
                ids = [sid, 'local_id', id, cid]
            else:
                ids = [sid, 'single', id]
        if ids is not None:
            obj = xf.gets(self.singles, ids)
            if obj is not None:
                #raise IOCError(f"null for {ids}")
                return obj
        source = xf.g(data, source=0)
        if source == 0:
            raise Exception(f"define object without 'source' key, {data}")
        fc = xf.get(self.sources, source, None)
        if fc is None:
            fc = pyz.load(source)
            self.sources[source]=fc
        cst = xf.g(data, construct = None)
        if cst is None:
            _args = xf.g(data, args = [])
            _maps = xf.g(data, maps = {})
            cst = [_args, _maps]
        cst = self.fmt_cst(cst)
        args = xf.g(cst, args=[])
        maps = xf.g(cst, maps={})
        args = [self.get_obj(v, conf) for v in args]
        maps = {k:self.get_obj(maps[k], conf) for k in maps}
        obj = fc(*args, **maps)
        if ids is not None:
            xf.sets(self.singles, ids, obj)
        prev_call = xf.g(data, prev_call=None)
        if prev_call is not None:
            # TODO: 这边info透传好像会有问题
            self.get_obj(prev_call, conf, obj, edata.info)
        sets = xf.g(data, sets=[])
        if type(sets)==list:
            for kv in sets:
                kv = self.fmt_set(kv)
                k = kv['key']
                v = xf.get_first(kv, "val", "data")
                v = self.get_obj(v, conf, obj, edata.info)
                setattr(obj, k, v)
        else:
            for k,v in sets.items():
                v = self.get_obj(v, conf, obj, edata.info)
                setattr(obj, k, v)
        call = xf.g(data, call=None)
        if call is not None:
            self.get_obj(call, conf, obj, edata.info)
        return obj
    def remove(self, edata:EncapeData):
        sid = edata.sid
        data = edata.data
        data = self.format(data)
        info = edata.info
        conf = edata.conf
        confs = edata.confs
        if type(info) == dict:
            cid = xf.g(info, id=None)
        else:
            cid = None
        id = xf.g(data, id = None)
        single = xf.g(data, single=None)
        if id is None:
            single = 0
        ids = None
        if single or cid is not None:
            if single:
                ids = [sid, id]
            else:
                ids = [sid, id, cid]
        if ids is None:
            return None
        obj = xf.gets(self.singles, ids)
        if obj is None:
            return None
        call = xf.g(data, remove=None)
        if call is not None:
            self.get_obj(call, conf, obj, edata.info)
        xf.removes(self.singles, ids)
        call = xf.g(data, after_remove=None)
        if call is not None:
            self.get_obj(call, conf, obj, edata.info)
        return None

pass



#coding=utf-8
import os
"""
读写文件再简化
"""
def fread(fp, mode='rb'):
    with open(fp, mode) as f:
        return f.read()

pass
read=fread
def freads(fp, mode = 'rb', size=1024*1024):
    with open(fp, mode) as f:
        while True:
            bs = f.read(size)
            if len(bs)==0:
                break
            yield bs

pass
reads=freads

def fwrite(ct, fp, mode = 'wb'):
    with open(fp, mode) as f:
        f.write(ct)

pass
write = fwrite
def fwrites(cts, fp, mode = 'wb'):
    with open(fp, mode) as f:
        for ct in cts:
            f.write(ct)

pass
writes = fwrites

def makedir(dp):
    if os.path.isdir(dp):
        return
    os.makedirs(dp)

pass
def makefdir(fp):
    fp = os.path.abspath(fp)
    dp = os.path.dirname(fp)
    makedir(dp)

pass

def dirpath(fp, n=1):
    for i in range(n):
        fp = os.path.dirname(fp)
    return fp

pass

dirname = dirpath

def fcover(filepath, wsize = 1024*10):
	st = os.stat(filepath)
	size = st.st_size
	bs = b'a'*wsize
	with open(filepath, 'wb') as f:
		for i in range(0, size, wsize):
			f.write(bs)

pass
cover = fcover
def removes(fp, cover = False):
    if not os.path.exists(fp):
        return
    if os.path.isfile(fp):
        #print(f"remove file '{fp}'")
        if cover:
            fcover(fp)
        os.remove(fp)
        return
    fps = os.listdir(fp)
    fps = [os.path.join(fp, f) for f in fps]
    [removes(f, cover) for f in fps]
    #print(f"removedirs '{fp}'")
    os.rmdir(fp)

pass


from .. import base
from .. import item
from .. import exp
from . import lr
from . import reval
class Fcs:
    def __init__(self):
        self.maps = {}
    def set(self, _type, fc):
        self.maps[_type] = fc
    def __call__(self, val, _type):
        if _type not in self.maps:
            raise Exception("unreginize type:"+_type)
        return self.maps[_type](val)

pass


class LRValDeal(lr.LRDeal):
    """
    """
    def init(self, left, right, fc):
        super().init(left, right, 'lrval')
        self.fc = fc
    def build(self, arr):
        rst = []
        if len(arr)!=2:
            raise Exception("error in lrval:"+arr)
        for _item in arr:
            if not _item.is_val:
                raise Exception("error in list:"+_item)
            rst.append(_item.val)
        try:
            val = self.fc(rst[0], rst[1])
        except Exceptin as exp1:
            print("exp:", exp1)
            raise Exception(f"error in lrval fc: {self.fc}({rst}): {exp1} ")
        return item.Item(val, type='val', is_val = 1)
    def build_arr(self, obj):
        return self.build(obj)

pass



#python
"""
测试代码
测试xf.loads和json.loads的速度
在修改json的scanner.py(不让它用c_make_scanner这个底层优化函数)
耗时对比如下：

json
time cost in <function loads at 0x0000015DEDD0D9E0>: 0.04396796226501465

xf
time cost in <function loads at 0x0000015DEDD0F100>: 0.3307473659515381

慢7倍多，感觉还能接受，之前没注意速度，更早的版本(buildz.xf.read.loads，后面可能会删掉)耗时要三四秒，对比之后就重新写了个，也不管什么堆栈了，就递归调用，后面有时间再改回堆栈吧（python的list的append和pop效率貌似不咋地，尤其是pop(0)）
"""
from buildz.xf import readz as rz
from buildz.xf import read as rd
from buildz import xf
import json
import time

def cost(n, f,*a,**b):
    c = time.time()
    r = f(*a,**b)
    d = time.time()-c
    print(f"time cost in {n}-{f}: {d}")
    return r

pass

n = 100
m = 12
l = 11
_arr = [123]
print("test A")
for i in range(n):
    _arr = [list(_arr)]

pass
print("test B")
_map = {}
for i in range(m):
    _map[i] = dict(_map)

pass
print("test C")
rst = []
for i in range(l):
    rst.append([_arr,_map])

pass
print("test D")
json.dumps(_arr)
print("test E")
json.dumps(_map)
print("test F")
js = json.dumps(rst)
#js = "\n\n"+js+"\n"
#js = xf.dumps(rst, json_format=1)
# js = r"""
# [
#     1,2,3,{"4":5,"6":7,"8":9,"10":11,"4":6}
# ]
# """
print("start")
jv = cost("json.loads", json.loads,js)
xv = cost("rz.loads",rz.loads,js)
print(f"judge: {jv==xv}")
#_xv = cost("rd.loads",rd.loads, js)
#with open("test.json", 'w') as f:
#    f.write(js)
if n>3 or m>3 or l > 3:
    exit()
print(json.dumps(jv))
print(json.dumps(xv))







import sys
import os
import hashlib
def hashcode(s):
    if type(s)==str:
        s = s.encode("utf-8")
    return hashlib.md5(s).hexdigest()

pass
def test_current(fp, up = 1):
    """
        将当前目录的上{up}层目录加入sys.path，这样可以在同层写测试代码，但是import还是要当在上{up}层做import
    """
    dp = os.path.dirname(os.path.abspath(fp))
    for i in range(up):
        dp = os.path.dirname(dp)
    sys.path.insert(0, dp)

pass
def load(md, fc = None):
    """
        import object(whether module or others) from md(or md.fc)
        exp:
            load("buildz.xf") = package xf
            load("buildz.xf", "loads") = function loads from package buildz.xf
            load("buildz.xf.loads") = function loads from package buildz.xf
    """
    mds = md.split(".")
    arr = mds[1:]
    while len(mds)>0:
        try:
            md = __import__(".".join(mds))
            break
        except ModuleNotFoundError as exp:
            mds = mds[:-1]
    if len(mds)==0:
        raise Exception("can't import package from "+md)
    for k in arr:
        md = getattr(md, k)
    if fc is not None:
        fc = getattr(md, fc)
    else:
        fc = md
    return fc

pass
def pyexe():
    return sys.executable

pass
exe=pyexe
is_windows = sys.platform.lower()=='win32'
def pypkg():
    """
        return python package path, test on linux and windows
    """
    import site
    sites = site.getsitepackages()
    if is_windows:
        fpath = sites[-1]
    else:
        fpath = sites[0]
    return fpath

pass
pkg = pypkg
pth = pypkg

class Pth:
    def __init__(self, fp = "build.pth"):
        self.fp = os.path.join(pth(), fp)
    def read(self):
        fp = self.fp
        if not os.path.isfile(fp):
            return []
        with open(fp, 'rb') as f:
            s = f.read().decode()
        return s.split("\n")
    def add(self, path):
        arr = self.read()
        if path in arr:
            print("alread add")
            return
        arr.append(path)
        self.write(arr)
    def write(self, paths = []):
        if type(paths) not in [list, tuple]:
            paths = [paths]
        s = "\n".join(paths)
        with open(self.fp, 'wb') as f:
            f.write(s.encode())
    def remove(self):
        os.remove(self.fp)

pass

_pth = Pth()



#coding=utf-8

from . import dirz

import re
class ListsDeal(dirz.FileDeal):
    def result(self):
        return self.fps, self.errors
    def init(self):
        super().init()
        self.fps = []
        self.errors = []
    def visit(self, finfo, depth):
        fp, isdir = finfo.path, finfo.isdir
        self.fps.append([fp, isdir])
        return True
    def catch(self, finfo, depth, exp):
        fp, isdir = finfo.path, finfo.isdir
        self.errors.append([fp, isdir, exp])

pass

class SearchDeal(dirz.FileDeal):
    def init(self, pt_fp=None, pt = None, pt_dp = None, depth = None, relative = False):
        super().init()
        self.pt_fp = pt_fp
        self.pt_dp = pt_dp
        if type(pt) == str:
            pt = pt.encode()
        self.pt = pt
        self.rst = []
        self.errs = []
        self.depth = depth
        self.relative = relative
    def fps(self, keep_dir = False, relative = None, fp_only = True):
        if relative is None:
            relative = self.relative
        rst = self.rst
        if relative:
            rst = [[i.rpath, i.isdir] for i in rst]
        else:
            rst = [[i.path, i.isdir] for i in rst]
        if not keep_dir:
            rst = [k for k in rst if not k[1]]
        if fp_only:
            rst = [k[0] for k in rst]
        return rst
    def result(self):
        return self.rst
    def visit(self, finfo, depth):
        if self.depth is not None and depth > self.depth:
            return False
        filepath = finfo.path
        isdir = finfo.isdir
        if self.relative:
            filepath = finfo.rpath
        if isdir:
            if finfo.empty_dir or depth==self.depth:
                if self.pt_dp is not None and len(re.findall(self.pt_dp, filepath))==0:
                    return True
                self.rst.append(finfo)
            return True
        if self.pt_fp is not None and len(re.findall(self.pt_fp, filepath))==0:
            return False
        if self.pt is not None:
            try:
                with open(filepath, 'rb') as f:
                    s = f.read()
            except Exception as exp:
                self.catch(finfo, depth, exp)
                return False
            if len(re.findall(self.pt, s))==0:
                return False
        self.rst.append(finfo)
        return False
    def catch(self, finfo, depth, exp):
        filepath = finfo.path
        if self.relative:
            filepath = finfo.rpath
        print(f"exp in {finfo}: {exp}")
        self.errs.append([finfo, exp])
        pass

pass

def lists(fp):
    return ListsDeal().dirs(fp)

pass
def _search(dp, pt_fp = None, pt = None, pt_dp = None, depth=None, relative = False):
    deal = SearchDeal(pt_fp, pt, pt_dp, depth,relative)
    return deal.dirs(dp)

pass
def searchs(dp, pt_fp = None, pt = None, pt_dp = None, depth=None, relative = False):
    deal = SearchDeal(pt_fp, pt, pt_dp, depth,relative)
    deal.dirs(dp)
    return deal.fps(keep_dir=True, fp_only=False)

pass
def searchd(dp, pt_fp = None, pt = None, pt_dp = None, depth=None, relative = False):
    deal = SearchDeal(pt_fp, pt, pt_dp, depth,relative)
    deal.dirs(dp)
    return deal.fps(keep_dir=True, fp_only=True)

pass

def search(dp, pt_fp = None, pt = None, depth=None, relative = False):
    deal = SearchDeal(pt_fp, pt, None, depth,relative)
    deal.dirs(dp)
    return deal.fps(keep_dir=False, fp_only=True)

pass


#

from .ioc import confs
from .ioc_deal.base import default_deals

def build(fp_init = None, add_default_deals = True):
    obj = confs.Confs()
    if (fp_init is not None):
        obj.init_fp(fp_init)
    if add_default_deals:
        obj.add_fp(default_deals)
    return obj

pass